
#include "QuasiNewton_ver2.h"

//=========================================================================
int quasi_Newton_method( double*      &x, 
			 double*      &f,
			 double**     &H, 
			 MULTI_FUN     func,
			 const int    &N, 
			 const double &eps )
{
  int flag=1;
  double *p  = new double[N];
  double *q  = new double[N];
  double *dx = new double[N];
  double inv_ss_sHF, ss=0.0, sHF=0.0;

  func( x, f, N );
  
  while(1){

    ///////////////////////////    
    // Convergence Criterion //
    double max = 0.0e0;
    for( int j=0; j<N; j++ ){ if( fabs(f[j]) > max ){ max = fabs(f[j]); } }
    //printf( "max=%e\n",max);
    if( max < eps ){
      printf( "max=%e\n", max);
      flag=0;
      break; 
    }
    ///////////////////////////

    for( int j=0; j<N; j++ ){ 
      dx[j] = 0.0;
      for( int i=0; i<N; i++ ){ dx[j] -= H[j][i]*f[i]; }
    }

    for( int j=0; j<N; j++ ){ x[j] += dx[j]; }

    func( x, f, N );
    double ss = 0.0, sHF = 0.0 ;
    for( int i=0; i<N; i++ ){ 
      ss += dx[i]*dx[i];
      q[i] = 0.0; p[i] = 0.0;
      for( int j=0; j<N; j++ ){
	q[i] +=       H[i][j]*f[j];
	p[i] += dx[j]*H[j][i];
	sHF  += dx[i]*H[i][j]*f[j];
      }
    }
    double inv__ss_plus_sHF = 1.0/(ss+sHF);
    for( int j=0; j<N; j++ ){ 
      for( int i=0; i<N; i++ ){ H[i][j] -= q[i]*p[j] * inv__ss_plus_sHF; }
    }
  }

  delete [] q; 
  delete [] p;   
  delete [] dx;
  return(flag);
}
//=========================================================================
//=========================================================================
//=========================================================================
//=========================================================================
// 
// (1) params : we can extract values of parameters by name
// (2) x[]    : variables f_i(x) = 0 (i=0,...,QUASI_NEWTON:N)
// (3) MULTI_FUNC_2 func : function which is zero for x, i.e., func = f in (2)
//      where int MULTI_FUN( IOParameterSet& params, double* x,
//                           double* f, const int &N );
//
//=========================================================================
//-------------------------------------------------------------------------
int quasi_Newton_method_2( IOParameterSet &params,
			   double*        &x,
			   double**       &H, 
			   MULTI_FUN_2     func
			   )
{
  int flag=1;
  static int N      = params.int_value("QUASI_NEWTON:N");
  static double eps = params    .value("QUASI_NEWTON:eps");
  static double *p  = new double[N];
  static double *q  = new double[N];
  static double *dx = new double[N];
  static double *f  = new double[N];

  //for(int i=0;i<N;i++){ for(int j=0;j<N;j++){ H[i][j]=0.0; }}
  //Jacobi_func( params, x, multi_func, H, N );
  //GJE0( H, N );

  func( params, x, f );

  while(1){

    ///////////////////////////    
    // Convergence Criterion //
    double max = 0.0e0;
    for( int j=0; j<N; j++ ){ if( fabs(f[j]) > max ){ max = fabs(f[j]); } }
    printf( "max=%e\n",max);
    if( max < eps ){
      printf( "max=%e\n", max);
      flag=0;
      break; 
    }
    ///////////////////////////
    
    for( int j=0; j<N; j++ ){ 
      dx[j] = 0.0;
      for( int i=0; i<N; i++ ){ 
	dx[j] -= H[j][i]*f[i];
      }
    }

    double min = 1.0e10;
    for( int j=0; j<N; j++ ){ 
      x[j] += dx[j];
      if( fabs(dx[j]) < min ){ min = fabs(dx[j]); }
    }
    if( min < eps ){ 
      printf("\n *********************** \n");
      for( int j=0; j<N; j++ ){	printf("x[%d]=%e \t", j, x[j]);  }
      printf("\n");
      for( int j=0; j<N; j++ ){ printf("f[%d]=%e \t", j, f[j] ); }
      printf("\n *********************** \n");
      printf("This is solution ! for eps = %e \n", eps);
      return(0);
    }

    func( params, x, f );

    double ss = 0.0, sHF = 0.0 ;
    for( int i=0; i<N; i++ ){ 
      ss += dx[i]*dx[i];
      for( int j=0; j<N; j++ ){
	sHF += dx[i]*H[i][j]*f[j];
      }
    }

    double inv__ss_plus_sHF = 1.0/(ss+sHF);    

    //printf("ss =%e \n", ss  );
    //printf("sHF=%e \n", sHF );
    //printf("inv__ss_plus_sHF = %e \n", inv__ss_plus_sHF );

    for( int i=0; i<N; i++ ){ 
      q[i] = 0.0; p[i] = 0.0;
      for( int j=0; j<N; j++ ){
	q[i] +=       H[i][j]*f[j];
	p[i] += dx[j]*H[j][i];
      }
    }

    for( int j=0; j<N; j++ ){ 
      for( int i=0; i<N; i++ ){ 
	H[i][j] += - q[i]*p[j] * inv__ss_plus_sHF;
      }
    }

  }

  return(flag);
}
//=========================================================================
int Jacobi_Matrix_2( IOParameterSet &params,
		     double*       &x0,
		     MULTI_FUN_2    func,
		     double**      &J
		   )
{
  static int N  = params.int_value("QUASI_NEWTON:N");  
  double dx     = 1.0e-3;
  double inv_dx = 1.0/dx;
  //J[i][j] // i: runs for functions, j: runs for variables

  static double *x1 = new double[N];
  static double *fo = new double[N];
  static double *fd = new double[N];

  func( params, x0, fo );

  for( int j=0; j<N; j++ ){

    // delta(,): Kronecker's delta
    for( int jj=0; jj<N; jj++ ){ x1[jj]=x0[jj]+delta(j,jj)*dx; }

    func( params, x1, fd );
    
    for( int i=0; i<N; i++ ){
      J[i][j] = (fd[i]-fo[i])*inv_dx;
      //printf("J[%d][%d]=%e \n", i, j, J[i][j] );
    }
  }
  return(0);
}
//=========================================================================
